local addonName, addon = ...
AvengersNemesis = addon
SLASH_AVENGERSNEM1 = "/avn"
local avnMainFrameX = 300
local avnMainFrameY = 300
local avnMainFrameW = 335
local avnMainFrameH = 420
local numRaces = 7
local avnSubFrameNum = 8
local myName = UnitName("player")
local myFullName = myName.."-"..GetRealmName()
local groups
local factionGroup, factionName = UnitFactionGroup("player")
if factionGroup == 'Alliance' then
  groups = {"Orc","Blood Elf","Troll","Tauren","Goblin","Undead","Panda","None"}
elseif factionGroup == 'Horde' then
  groups = {"Human","Dwarf","Night Elf","Gnome","Worgen","Draenei","Panda","None"}
else
  groups = {}
end

local mt
local AVN_mainFrame = {}
local AVN_subFrame = {}
local ROW_HEIGHT = 14
local MAX_ROWS = 5   

local function printGroupsToRW()
 local rw = "|| "
 for i=1,numRaces do
  rw = rw..i..string.char(58).." "..groups[i].." || "
 end
 --SendChatMessage("Avengers Nemesis Groups:", "RAID_WARNING")
 SendChatMessage(string.format("If you are on a Nemesis quest, whisper %s with 'avn RACE' (eg: avn %s)",myFullName,groups[1]), "RAID_WARNING")
 --SendChatMessage(rw, "RAID_WARNING")
 SendChatMessage("For help whisper me 'avn ?'", "RAID_WARNING")
end

local caddonName = "\124cFF00FF00"..addonName.."\124r: "
local function chatMsg(...)
  print(caddonName..string.format(...))
end

addon.debug = false
local function debug(...)
  if addon.debug then
    chatMsg(...)
  end
end

local raceQuest = {
  [36921] = "Human", 
  [36960] = "Undead", 
  [36964] = "Orc", 
  [36962] = "Tauren", 
  [36934] = "Panda", 
  [36958] = "Blood Elf", 
  [36928] = "Worgen", 
  [36968] = "Panda", 
  [36966] = "Troll", 
  [36930] = "Draenei", 
  [36932] = "Night Elf", 
  [36970] = "Goblin", 
  [36926] = "Gnome", 
  [36924] = "Dwarf", 
}

local svnrev = select(3,string.find("$Revision: 31 $", ".* (.*) .*"))
addon.version = (GetAddOnMetadata("AvengersNemesis", "X-Curse-Packaged-Version") or
                 GetAddOnMetadata("AvengersNemesis", "Version") or "@")
if string.find(addon.version, "@") then -- dev copy uses "@.project-version.@"
  addon.version = "r"..svnrev
  addon.debug = true
elseif not addon.version:find("^r") then
  addon.version = "v"..addon.version
end


function addon:normalizeRace(race) -- normalized spelling of the race
  return race:gsub("[sS]$",""):gsub("[%s%p]*",""):upper()
end

function addon:shortenRace(race) -- initial substring
  return addon:normalizeRace(race):match("^%s*(%w%w)")
end

function addon:raceToGroup(race) -- accepts race name or group number and returns valid group number or nil
  if not addon._r2g then
    addon._r2g = {}
    for id,race in ipairs(groups) do
      addon._r2g[addon:normalizeRace(race)] = id
      addon._r2g[addon:shortenRace(race)] = id
    end
    for a,r in pairs({
      -- other abbreviations
      NE = 'Night Elf',
      BE = 'Blood Elf',
      Forsaken = 'Undead',
      Kaldorei = 'Night Elf',
      Sindorei = 'Blood Elf',
      Tushui = 'Panda',
      Huojin = 'Panda',
      Kezan = 'Goblin',
    }) do
      addon._r2g[addon:normalizeRace(a)] = addon._r2g[addon:normalizeRace(r)]
    end
  end
  if not race or #race == 0 then return nil end
  local group = tonumber(race)
  if group and group >= 1 and group <= 8 and math.floor(group) == group then
    return group -- numerical parse
  else
    group = addon._r2g[addon:normalizeRace(race)] -- race parse
    if group then return group end
    -- either their spelling sucks or they abbreviated, try a best-effort initial substring match
    return  addon._r2g[addon:shortenRace(race)]
  end
end

function addon:colorUnit(unit)
  local _,class = UnitClass(unit)
  local color = class and RAID_CLASS_COLORS[class]
  color = color and color.colorStr
  color = color or "ffcfcfcf"
  return string.format("\124c%s%s\124r",color,unit)
end

local function setAVNframeText()
 for group = 1, avnSubFrameNum do
   AVN_subFrame[group]:Update()
 end
 if addon.autobutton:GetChecked() then
   addon.autoFrame:Show() -- schedule auto-arrange
 end
end

local function clearAVNdata(alwaysshow)
    StaticPopupDialogs["AVN_PROMPTFORRESET"] = StaticPopupDialogs["AVN_PROMPTFORRESET"] or {
      text = "Do you want to clear the AVN data?",
      button1 = "Yes",
      button2 = "No",
      OnAccept = function()
        mt = {}
        setAVNframeText()
      end,
      OnCancel = function()
      end,
      timeout = 0,
      whileDead = true,
      hideOnEscape = true,
      preferredIndex = 3,
    }
    local now = GetTime()
    if now < (addon.lastclearprompt or 0)+5*60 and not alwaysshow then return end -- don't spam messages
    addon.lastclearprompt = now
    StaticPopup_Show ("AVN_PROMPTFORRESET")
end

local function ToggleDock()
  addon.docked = not addon.docked
  addon.dockbutton:SetText(addon.docked and "Undock" or "Dock")
  if addon.docked then
    --OpenFriendsFrame(4) -- causes taint :-(
    AVN_mainFrame:SetFrameLevel(RaidFrame and RaidFrame:GetFrameLevel()+10 or 50)
  end
  hookRaidFrame()
end
function hookRaidFrame()
  if not RaidFrame then return end
  if addon.raidhooked then return end
  if IsAddOnLoaded("Blizzard_RaidUI") then
    addon.raidhooked = true
  end
  RaidFrame:HookScript("OnShow", function()
    if addon.docked and addon:shouldIshow(false) then
      AVN_mainFrame:Show()
    end
  end)
  RaidFrame:HookScript("OnHide", function()
    if addon.docked then
      AVN_mainFrame:Hide()
    end
  end)
end
hookRaidFrame()

local contextFrame = CreateFrame("Frame", "AVNUserContextFrame", UIParent, "UIDropDownMenuTemplate")
local contextRow
local contextMenu = {
    { text = "Title", isTitle = true},
    { text = "Whisper", func = function() ChatFrame_SendTell(contextRow.name) end },
    { text = "Send Help", func = function() addon:sendHelpFile(contextRow.name) end },
    { text = "Add to Queue", hasArrow = true,
        menuList = { } -- filled in below
    },
    { text = "Remove from Queue", func = function() addon:removeFromGroup(contextRow.name, true) end },
    { text = "Close", func = function() contextFrame:Hide() end },
}
for i=1,numRaces do
  contextMenu[4].menuList[i] = 
     { text = string.format("%d: %s",i,groups[i]), 
       func = function() 
	 contextFrame:Hide() -- hide top-level menu
         addon:addToGroup(contextRow.name, i, true)
       end }
end
local function Row_OnClick(self)
  contextRow = self
  debug("Context menu: "..contextRow.name)
  contextMenu[1].text = self:GetText()
  contextFrame:Show()
  EasyMenu(contextMenu, contextFrame, "cursor", 0 , 0, "MENU", 10)
end

local function registerHelp(frame, text)
  frame:SetScript("OnEnter", function() 
     GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
     GameTooltip:SetText(text) 
     GameTooltip:Show()
  end)
  frame:SetScript("OnLeave", function() GameTooltip:Hide() end)
end

local function drawAvnMainFrame()
    AVN_mainFrame = CreateFrame("Frame", "AVN_MainFrame", UIParent)
    AVN_mainFrame:SetMovable(true)
    AVN_mainFrame:EnableMouse(true)
    AVN_mainFrame:RegisterForDrag("LeftButton")
    AVN_mainFrame:SetScript("OnDragStart", AVN_mainFrame.StartMoving)
    AVN_mainFrame:SetScript("OnDragStop", AVN_mainFrame.StopMovingOrSizing)
    AVN_mainFrame:SetSize(avnMainFrameW,avnMainFrameH)
    AVN_mainFrame:SetClampedToScreen(true)
    AVN_mainFrame:SetPoint("TOPLEFT",400,-130)
    AVN_mainFrame.backGround=AVN_mainFrame:CreateTexture("AVN_mainFrame_backGround","BACKGROUND")
    AVN_mainFrame.backGround:SetSize(avnMainFrameW,avnMainFrameH)
    AVN_mainFrame.backGround:SetPoint("CENTER")
    AVN_mainFrame.backGround:SetTexture(0,0,0,.5)
    AVN_mainFrame.title = AVN_mainFrame:CreateFontString(AVN_mainFrame_Title, "OVERLAY", "GameFontNormal")
    AVN_mainFrame.title:SetPoint("TOP", 0, -4)
    AVN_mainFrame.title:SetText("Avengers Nemesis Groups "..addon.version)
    AVN_mainFrame:SetScript("OnShow", function() 
      ToggleDock() -- enable docking on first show
      AVN_mainFrame:SetScript("OnShow", nil)
    end)

    local buttonBackdrop = {bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
        edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
        tile = true, tileSize = 16, edgeSize = 16, 
        insets = { left = 4, right = 4, top = 4, bottom = 4 }}

    local closeFrame = CreateFrame("Button", nil, AVN_mainFrame)
    closeFrame:SetSize(22, 20)
    closeFrame:SetNormalFontObject(GameFontHighlightCenter)
    closeFrame:SetPoint("CENTER", AVN_mainFrame, "TOPRIGHT", -13, -12)
    closeFrame:SetText("X")
    closeFrame:SetBackdrop(buttonBackdrop)
    closeFrame:SetBackdropColor(1,0,0,1);
    closeFrame:SetScript("OnClick", function(self, arg1)
      AVN_mainFrame:Hide()
    end)
    registerHelp(closeFrame, "Close this display and work in background")

    local printRaid = CreateFrame("Button", nil, AVN_mainFrame)
    printRaid:SetSize(40, 20)
    printRaid:SetNormalFontObject(GameFontHighlightCenter)
    printRaid:SetPoint("CENTER", AVN_mainFrame, "BOTTOMRIGHT", -22, 12)
    printRaid:SetText("RW")
    printRaid:SetBackdrop(buttonBackdrop)
    printRaid:SetBackdropColor(0,0,0,.3);
    printRaid:SetScript("OnClick", function(self, arg1)
      printGroupsToRW()
    end)
    registerHelp(printRaid, "Send a raid warning with instructions")

    local clearData = CreateFrame("Button", nil, AVN_mainFrame)
    clearData:SetSize(75, 20)
    clearData:SetNormalFontObject(GameFontHighlightCenter)
    clearData:SetPoint("CENTER", AVN_mainFrame, "BOTTOMLEFT", 40, 12)
    clearData:SetText("Clear Data")
    clearData:SetBackdrop(buttonBackdrop)
    clearData:SetBackdropColor(0,0,0,.3);
    clearData:SetScript("OnClick", function(self, arg1)
      clearAVNdata(true)
    end)
    registerHelp(clearData, "Clear all character data")

    local dock = CreateFrame("Button", nil, AVN_mainFrame)
    addon.dockbutton = dock
    dock:SetSize(55, 20)
    dock:SetNormalFontObject(GameFontHighlightCenter)
    dock:SetPoint("CENTER", AVN_mainFrame, "TOPLEFT", 30, -12)
    dock:SetText("Dock")
    dock:SetBackdrop(buttonBackdrop)
    dock:SetBackdropColor(0,0,0,.3);
    dock:SetScript("OnClick", ToggleDock)
    registerHelp(dock, "Dock display to Raid frame")

    local auto = CreateFrame("CheckButton", "AVN_autoArrange", AVN_mainFrame, "UICheckButtonTemplate")
    addon.autobutton = auto
    auto:SetSize(20, 20)
    _G[auto:GetName().."Text"]:SetText("Auto-Arrange")
    auto:SetPoint("CENTER", AVN_mainFrame, "BOTTOM", -35, 12)
    auto:SetChecked(true)
    auto:SetScript("OnClick",function() addon.autoFrame:Show() end)
    registerHelp(auto, "Automatically arrange players into raid groups")

--    AVN_mainFrame:Show()
    AVN_mainFrame:Hide()
end
local function drawAvnSubBoxes()
    local boxW = 150
    local boxH = 71
    local spacer = 20

    for i = 1, avnSubFrameNum do
        AVN_subFrame[i] = CreateFrame("Frame", "AVN_MainFrame", AVN_mainFrame)
        AVN_subFrame[i]:SetSize(boxW, boxH)
        if i == 1 then
            AVN_subFrame[i]:SetPoint("TOPLEFT", 15, -45)
        elseif i == 3 or i == 5 or i == 7 then
            AVN_subFrame[i]:SetPoint("TOPLEFT", AVN_subFrame[i-2], "BOTTOMLEFT", 0, -spacer)
        else
            AVN_subFrame[i]:SetPoint("TOPLEFT", AVN_subFrame[i-1], "TOPRIGHT", 15, 0)
        end
        AVN_subFrame[i].backGround=AVN_subFrame[i]:CreateTexture("AVN_subFrame_backGround","BACKGROUND")
        AVN_subFrame[i].backGround:SetSize(boxW, boxH)
        AVN_subFrame[i].backGround:SetPoint("CENTER")
        AVN_subFrame[i].backGround:SetTexture(1,1,1,.2)
        AVN_subFrame[i].title = AVN_subFrame[i]:CreateFontString(AVN_subFrame_Title, "OVERLAY", "GameFontNormal")
        AVN_subFrame[i].title:SetPoint("TOPLEFT", -10, 14)
        AVN_subFrame[i].title:SetText(i.." "..groups[i])

        local bar = CreateFrame("ScrollFrame", "$parentScrollBar"..i, AVN_subFrame[i], "FauxScrollFrameTemplate")
        bar:SetPoint("TOPLEFT", 0, -1)
        bar:SetPoint("BOTTOMRIGHT", -23, 1)
        bar:SetScript("OnVerticalScroll", function(self, offset)
          self.offset = math.floor(offset / ROW_HEIGHT + 0.5)
          AVN_subFrame[i]:Update()
        end)
        bar:SetScript("OnShow", function()
        	AVN_subFrame[i]:Update()
        end)
        AVN_subFrame[i].scrollBar = bar

        AVN_subFrame[i].Update = function()
	  AVN_subFrame[i].data = addon:getSortedMembers(i, AVN_subFrame[i].data)
          local myGroupData = AVN_subFrame[i].data
	  local count = #myGroupData
          FauxScrollFrame_Update(AVN_subFrame[i].scrollBar, count, MAX_ROWS, ROW_HEIGHT)
          local offset = FauxScrollFrame_GetOffset(AVN_subFrame[i].scrollBar)
          for j = 1, MAX_ROWS do
            local row = AVN_subFrame[i].rows[j]
            local id = j + offset
            if id <= count then
	      local name = myGroupData[id]
	      local text = addon:colorUnit(name)
	      local curgroup = addon.groupcache and addon.groupcache[name]

	      if IsInRaid() then
	        if not UnitIsConnected(name) then -- disconnected
	          row.highlight:SetTexture(0,0,0,0.4)
		  text = "(DC)"..text
	        elseif not UnitInRaid(name) then -- not in raid, should never happen
	          row.highlight:SetTexture(0,0,0,0.7)
	        elseif curgroup and curgroup ~= i then -- unit not in correct group
	          row.highlight:SetTexture(1,0,0,0.5)
	          text = "("..curgroup..")"..text
		else -- normal
	          row.highlight:SetTexture(0,0,0,0)
	        end
	      end
              row:SetText(text)
	      row.name = name
              row:Show()
            else
              row:Hide()
            end
          end
        end

        local rows = setmetatable({}, { __index = function(t, j)
	    local row = CreateFrame("Button", "$parentRow"..i, AVN_subFrame[i])
            row:SetSize(146, ROW_HEIGHT)
            row:SetNormalFontObject(GameFontHighlightLeft)
	    row.highlight = row:CreateTexture(nil,"BACKGROUND")
	    row.highlight:SetSize(row:GetSize())
	    row.highlight:SetPoint("CENTER")
	    row.highlight:SetTexture(0,0,0,0)
	    row.highlight:Show()
            if j == 1 then
                row:SetPoint("TOPLEFT", AVN_subFrame[i], 2, -2)
            else
                row:SetPoint("TOPLEFT", AVN_subFrame[i].rows[j-1], "BOTTOMLEFT")
            end
	    row:SetScript("OnClick", Row_OnClick)
	    row:RegisterForClicks("AnyDown")
            rawset(t, j, row)
            return row
        end })
        AVN_subFrame[i].rows = rows

        AVN_subFrame[i]:Show()
    end
end

function addon:getSortedMembers(group, tableopt)
 local t = wipe(tableopt or {})
 for name,info in pairs(mt) do
  if info[1] == group then
   table.insert(t, name)
  end
 end
 addon.sortMemFunc = addon.sortMemFunc or function(a,b) return mt[a][2] < mt[b][2] end
 table.sort(t, addon.sortMemFunc)
 return t
end

function addon:getMembersOfGroup(group)
 local rText = ""
 local sortedKeys = addon:getSortedMembers(group)
 for _, key in ipairs(sortedKeys) do
  rText = rText..addon:colorUnit(key).."\n" 
 end
 return rText
end

addon.autoFrame = CreateFrame("Button",nil,UIParent)
addon.autoFrame:SetScript("OnUpdate", function (self, elap)
  addon.autoFrameElap = (addon.autoFrameElap or 0) + elap
  if addon.autoFrameElap < 1 then return end
  addon.autoFrameElap = 0
  if not addon.autobutton:GetChecked() then
    addon.autoFrame:Hide()
  end
  addon:autoArrange()
end)

function addon:autoArrange()
  if not addon:shouldIshow(false) or not addon.autobutton:GetChecked() then return end
  local now = GetTime()
  local movetimeout = 5
  addon.autoLastMove = addon.autoLastMove or {} -- persistent tracking of last move attempt time on a player
  addon.autoInsert = wipe(addon.autoInsert or {})
  -- find each group's highest-priority insertion candidate
  for group = 1,numRaces do
    for cnt,name in ipairs(AVN_subFrame[group].data) do
      if cnt > 5 then break end -- group is full
      local curgroup = addon.groupcache and addon.groupcache[name]
      local id = UnitInRaid(name)
      if curgroup and curgroup ~= group and id then 
        addon.autoInsert[id] = group
        break
      end
    end
  end
  -- find a victim for each group
  -- prioritize players with no assignment to reduce overall moves
  -- also prioritize least-recently moved players to ensure the algorithm doesn't get stuck if one victim is bugged and unmovable
  addon.autoVictim = wipe(addon.autoVictim or {}) -- each group's selected victim id
  addon.autoVictimPri = wipe(addon.autoVictimPri or {}) -- his victim priority
  for name, subgroup in pairs(addon.groupcache) do
    local id = UnitInRaid(name)
    if id then
      local lastmove = addon.autoLastMove[name] or 0
      local priority
      if not mt[name] then -- first priority: unassigned players
        priority = lastmove
      elseif mt[name][1] ~= subgroup  then -- second priority: players in wrong group
        priority = lastmove + (movetimeout * 2)
      end
      local bestpri = addon.autoVictimPri[subgroup]
      if priority and (not bestpri or priority < bestpri) then -- new winner
        addon.autoVictim[subgroup] = id
        addon.autoVictimPri[subgroup] = priority
      end
    end
  end
  -- select a move action
  -- for safety, only perform a single move at a time and don't attempt to move a player twice in a short time
  -- the move itself will trigger a GROUP_ROSTER_UPDATE that brings us back here for the next move, if needed
  local morework
  for id,group in pairs(addon.autoInsert) do
    local name = GetUnitName("raid"..id,true)
    if name and now > (addon.autoLastMove[name] or 0) + movetimeout then
      local vid = addon.autoVictim[group]
      if not vid then -- destination group empty
        debug("SetRaidSubgroup() %s (raid%d) to group %d",name,id,group)
	addon.autoLastMove[name] = now
        SetRaidSubgroup(id, group)
	return
      else
        local vname = GetUnitName("raid"..vid,true)
        if vname and now > (addon.autoLastMove[vname] or 0) + movetimeout then
          debug("SwapRaidSubgroup() %s (raid%d) with %s (raid%d) into group %d",name,id,vname,vid,group)
	  addon.autoLastMove[name] = now
	  addon.autoLastMove[vname] = now
	  SwapRaidSubgroup(id, vid)
	  return
	else
	  morework = true
        end
      end
    else
      morework = true
    end
  end
  if not morework then -- found no work, deschedule
    addon.autoFrame:Hide()
  end
end

function addon:addToGroup(name, group, quietErr)
 name = name:gsub("%s*-%s*"..GetRealmName(),"") -- strip realm for same-realm
 if UnitInRaid(name) then
  local old = mt[name]
  if old and old[1] == group then -- already in that group, don't lose timestamp
    if not quietErr then
       SendChatMessage("You are already in queue for group "..group.."-"..groups[group], "WHISPER", nil, name)
    end
    return
  end
  mt[name] = {group, time()}
  SendChatMessage("You have been added to the queue for group "..group.."-"..groups[group], "WHISPER", nil, name)
  SendChatMessage("Please whisper me with 'avn remove' when you have finished your quest or will be leaving the raid to make room for the next person in queue.", "WHISPER", nil, name)
  setAVNframeText()
 else
  if not quietErr then
   SendChatMessage("You are not in my raid. Please verify that you are sending your Nemesis group request to the correct person. If you sent this to me to be funny... ha... ha... ha...", "WHISPER", nil, name)
  end
 end
end

function addon:removeFromGroup(name, quiet)
 name = name:gsub("%s*-%s*"..GetRealmName(),"") -- strip realm for same-realm
 if mt[name] then
  local group = mt[name][1]
  mt[name] = nil
  if not quiet then
   SendChatMessage("You have been removed from the Nemesis Group queue. Thank you for joining Avengers Nemesis. Please come again.", "WHISPER", nil, name)
  end
  setAVNframeText()
 else
  if not quiet then
   SendChatMessage("You are not currently queued for a group.", "WHISPER", nil, name)
  end
 end
end

local function checkGroup(name)
 name = name:gsub("%s*-%s*"..GetRealmName(),"") -- strip realm for same-realm
 if mt[name] ~= nil then
  SendChatMessage("You are in queue for group "..mt[name][1].."-"..groups[mt[name][1]], "WHISPER", nil, name)
 else
  SendChatMessage("You are currently not in a Nemesis Group", "WHISPER", nil, name)
 end
end

local function getCountOfGroup(group, name)
 local count = 0
 for k,v in pairs(mt) do
  if v[1] == group then
   count = count + 1
  end
 end
 SendChatMessage(group..": "..groups[group].." - Total Current Members: "..count, "WHISPER", nil, name)
end

function addon:getGroupReport()
  local count = {}
  for name,info in pairs(mt) do
    local group = info[1] or 0
    count[group] = (count[group] or 0) + 1
  end
  local str = ""
  for i=1,numRaces do
    str = str..string.format("%s%s: %d", (#str>0 and ", " or ""), groups[i], count[i] or 0)
  end
  return "Nemesis group counts: "..str
end

function addon:sendHelpFile(name)
 SendChatMessage("Avengers Nemesis Groups "..addon.version.." Help:", "WHISPER", nil, name)
 SendChatMessage("'avn #' - Adds you to Nemesis group #", "WHISPER", nil, name)
 SendChatMessage("'avn list' - Replies with Races with group number", "WHISPER", nil, name)
 SendChatMessage("'avn check' - Replies with which group you are in", "WHISPER", nil, name)
 SendChatMessage("'avn count #' - Replies with how many people in Nemesis group #", "WHISPER", nil, name)
 SendChatMessage("'avn report' - Replies with a Nemesis group report", "WHISPER", nil, name)
 SendChatMessage("'avn remove' - Removes you from all Nemesis groups", "WHISPER", nil, name)
 SendChatMessage("'avn ?' - Replies with this information", "WHISPER", nil, name)
end

function addon:shouldIshow(output)
 if IsInRaid() then
  local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(UnitInRaid("player"))
  if rank > 0 then
   return true
  else
   if output then
    chatMsg("You are currently not in or raid or do not have assist")
   end
  end
 else
  if output then
   chatMsg("You are currently not in or raid or do not have assist")
  end
  return false
 end
end

function addon:scanGroups() -- populate groupcache with current party information
  if not IsInRaid() then return end
  addon.groupcache = wipe(addon.groupcache or {})
  for i=1,GetNumGroupMembers() do
    local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(i)
    if name and subgroup then
      addon.groupcache[name] = subgroup
    end
  end 
end

local function raidChangedCheckRoster()
  if not addon:shouldIshow(false) and next(mt) ~= nil then
    clearAVNdata()
    AVN_mainFrame:Hide()
  else
    for name,v in pairs(mt) do
      if not UnitInRaid(name) then
        addon:removeFromGroup(name, false)
      end
    end
  end
end

local function tellMeGroups(name)
 local rw = "|| "
 for i=1,#groups do
  rw = rw..i..string.char(58).." "..groups[i].." || "
 end
 SendChatMessage(rw, "WHISPER", nil, name)
end

-- Whisper message handler
local function getCommandType(msg, name)
 local qid = msg:match("\124Hquest:(%d+):") -- quest link means add
 local race = raceQuest[tonumber(qid)]
 if race then
   addon:addToGroup(name, addon:raceToGroup(race), false)
   return
 end
 local msg1, sep, msg2 = msg:match('^%s*(%a%a%a)([%s%p]*)(.-)%s*$')
 if msg1 and msg2 and msg1:lower() == 'avn' then
  msg2 = msg2:gsub("^add%s*","")
  local group = addon:raceToGroup(msg2)
  if group then
   addon:addToGroup(name, group, false)
   return
  end
  local comm, rest = msg2:match("^(%S*)%s*(.-)$")
  comm = comm and comm:lower()
  local group = addon:raceToGroup(rest)
  if comm == "list" then
   tellMeGroups(name)
  elseif comm == "check" then
   checkGroup(name)
  elseif comm == "count" and group then
   getCountOfGroup(group, name)
  elseif comm == "report" or comm == "count" then
    local str = addon:getGroupReport()
    SendChatMessage(str, "WHISPER", nil, name)
  elseif comm == "remove" then
   addon:removeFromGroup(name, false)
  elseif sep:trim() == "?" or comm == "help" then
   addon:sendHelpFile(name)
  else
   SendChatMessage("* AVN:Invalid request. Please reply with 'avn ?' for help.", "WHISPER", nil, name)
  end  
 end
end

local function innitalizeAVN()
    drawAvnMainFrame()
end

innitalizeAVN()
drawAvnSubBoxes()
addon:scanGroups()

AVN_mainFrame:RegisterEvent("VARIABLES_LOADED")
AVN_mainFrame:RegisterEvent("ADDON_LOADED")
AVN_mainFrame:RegisterEvent("CHAT_MSG_WHISPER")
AVN_mainFrame:RegisterEvent("PLAYER_LOGOUT")
AVN_mainFrame:RegisterEvent("GROUP_ROSTER_UPDATE")
AVN_mainFrame:RegisterEvent("PLAYER_ENTERING_WORLD")

function AVN_mainFrame:OnEvent(event, ...)
 local arg1, arg2 = ...
 if event == "ADDON_LOADED" and arg1 == "AvengersNemesis" then
  if AvengersNemesisCache == nil then
   mt = {}
   debug("no cache")
  else
   mt = AvengersNemesisCache
   debug("Got data cache")
  end
  setAVNframeText()
 elseif event == "PLAYER_LOGOUT" then
  AvengersNemesisCache = mt
 elseif event == "CHAT_MSG_WHISPER" then
  if addon:shouldIshow(false) then
   local a,b,c,d,e,f,g,h,i,j,k,l = ...
   getCommandType(a, b)
  elseif arg1 and arg1:match("^%s*[aA][vV][nN]") then
   SendChatMessage("I am not leading Nemesis Groups in this Avengers group", "WHISPER", nil, arg2)
  end
 elseif event == "GROUP_ROSTER_UPDATE" then
  raidChangedCheckRoster()
  addon:scanGroups()
  setAVNframeText()
 end
end
AVN_mainFrame:SetScript("OnEvent", AVN_mainFrame.OnEvent)

local function printAVNslashCommands()
 chatMsg("Avengers Nemesis "..addon.version.." Help:")
 chatMsg("/avn show - Show GUI")
 chatMsg("/avn hide - Hide GUI")
 chatMsg("/avn add # - Adds Target to group #")
 chatMsg("/avn remove - Removes Target from group")
 chatMsg("/avn count # - Replies with count of that group")
 chatMsg("/avn rw - Prints Raid Warning")
 chatMsg("/avn report - Send group report to raid")
 chatMsg("/avn clear - Clears all groups")
 chatMsg("/avn demo - Displays AVN frame regardless of raid state")
 chatMsg("/avn print - Prints queue array to main chat window")
 chatMsg("/avn help - Sends help file @Target")
 chatMsg("/avn ? - Displays this help file")
end
-- Slash Commands handler
local function displayAVNbox(msg, editbox)
    local command, rest = msg:match("^(%S*)%s*(.-)$")
    command = command and command:lower()
    local group = addon:raceToGroup(rest)
    if command == 'show' then
      if addon:shouldIshow(true) then
        AVN_mainFrame:Show()
      end
    elseif command == 'hide' then
        AVN_mainFrame:Hide()
    elseif command == 'add' and group then
      local who = ""
      if UnitExists("target") then
        who = GetUnitName("target", true)
      else
        who = myName
      end
      addon:addToGroup(who, group, true)
    elseif command == 'remove' then
      if UnitExists("target") then
        addon:removeFromGroup(GetUnitName("target", true), true)
      end
    elseif command == 'count' and group then
      getCountOfGroup(group, myName)
    elseif command == 'report' or command == 'count' then
      local str = addon:getGroupReport()
      if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
        SendChatMessage(str,"INSTANCE_CHAT")
      elseif IsInRaid() then
        SendChatMessage(str,"RAID")
      elseif IsInGroup() then
        SendChatMessage(str,"PARTY")
      else
        chatMsg(str)
      end
    elseif command == 'rw' then
        printGroupsToRW()
    elseif command == 'clear' then
        mt = {}
        setAVNframeText()
    elseif command == "demo" then
        AVN_mainFrame:Show()
    elseif command == "print" then
        for key,v in pairs(mt) do
          chatMsg(addon:colorUnit(key).." "..v[1].." "..v[2])
        end
    elseif command == "help" then
      if UnitExists("target") then
        addon:sendHelpFile(GetUnitName("target", true))
      end
    elseif command == "list" then
     tellMeGroups(GetUnitName("player", true))
    else
        debug(command.."|"..rest)
        printAVNslashCommands()
    end
end
SlashCmdList["AVENGERSNEM"] = displayAVNbox

